from __future__ import annotations

from tools.python_api.test.type_aliases import ConnDB

# required by python-lint


def test_param_empty(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    lst = [[]]
    conn.execute("CREATE NODE TABLE tab(id SERIAL, lst INT64[][], PRIMARY KEY(id))")
    result = conn.execute("CREATE (t:tab {lst: $1}) RETURN t.*", {"1": lst})
    assert result.has_next()
    assert result.get_next() == [0, lst]
    assert not result.has_next()
    result.close()


def test_issue_2874(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    result = conn.execute("UNWIND $idList as tid MATCH (t:person {ID: tid}) RETURN t.fName;", {"idList": [1, 2, 3]})
    assert result.has_next()
    assert result.get_next() == ["Bob"]
    assert result.has_next()
    assert result.get_next() == ["Carol"]
    assert not result.has_next()
    result.close()


def test_issue_2906(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    result = conn.execute("MATCH (a:person) WHERE $1 > a.ID AND $1 < a.age / 5 RETURN a.fName;", {"1": 6})
    assert result.has_next()
    assert result.get_next() == ["Alice"]
    assert result.has_next()
    assert result.get_next() == ["Carol"]
    assert not result.has_next()
    result.close()


def test_issue_3135(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    conn.execute("CREATE NODE TABLE t1(id SERIAL, number INT32, PRIMARY KEY(id));")
    conn.execute("CREATE (:t1 {number: $1})", {"1": 2})
    result = conn.execute("MATCH (n:t1) RETURN n.number;")
    assert result.has_next()
    assert result.get_next() == [2]
    assert not result.has_next()
    result.close()


def test_empty_list2(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    conn.execute(
        """
        CREATE NODE TABLE SnapArtifactScan (
            artifact_name STRING,
            scan_columns STRING[],
            scan_filter STRING,
            scan_limit INT64,
            scan_id STRING,
            PRIMARY KEY(scan_id)
        )
        """
    )
    result = conn.execute(
        """
        MERGE (n:SnapArtifactScan { scan_id: $scan_id })
        SET n.artifact_name = $artifact_name, n.scan_columns = $scan_columns
        RETURN n.scan_id
    """,
        {
            "artifact_name": "taxi_zones",
            "scan_columns": [],
            "scan_id": "896de6b9c7b69fa2598def49e8c61de07949be374d229a82899c9c75994fad20",
        },
    )
    assert result.has_next()
    assert result.get_next() == ["896de6b9c7b69fa2598def49e8c61de07949be374d229a82899c9c75994fad20"]
    assert not result.has_next()
    result.close()


def test_empty_map(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    conn.execute(
        """
        CREATE NODE TABLE Test (
            id SERIAL,
            prop1 MAP(STRING, STRING),
            prop2 MAP(STRING, STRING[]),
            PRIMARY KEY(id)
        )
        """
    )
    result = conn.execute(
        """
            CREATE (n:Test {prop1: $prop1, prop2: $prop2})
            RETURN n.id, n.prop1, n.prop2
        """,
        {
            "prop1": {"key": [], "value": []},
            "prop2": {"key": ["a"], "value": [[]]},
        },
    )
    assert result.has_next()
    assert result.get_next() == [0, {}, {"a": []}]
    assert not result.has_next()
    result.close()


# TODO(Maxwell): check if we should change getCastCost() for the following test
# def test_issue_3248(conn_db_readwrite: ConnDB) -> None:
#     conn, db = conn_db_readwrite
#     # Define schema
#     conn.execute("CREATE NODE TABLE Item(id UINT64, item STRING, price DOUBLE, vector DOUBLE[2], PRIMARY KEY (id))")
#
#     # Add data
#     conn.execute("MERGE (a:Item {id: 1, item: 'apple', price: 2.0, vector: cast([3.1, 4.1], 'DOUBLE[2]')})")
#     conn.execute("MERGE (b:Item {id: 2, item: 'banana', price: 1.0, vector: cast([5.9, 26.5], 'DOUBLE[2]')})")
#
#     # Run similarity search
#     result = conn.execute("MATCH (a:Item) RETURN a.item, a.price,
#               array_cosine_similarity(a.vector, [6.0, 25.0]) AS sim ORDER BY sim DESC")
#     assert result.has_next()
#     assert result.get_next() == [2]
#     assert result.has_next()
#     assert result.get_next() == [2]
#     assert not result.has_next()
#     result.close()



import os

import pexpect
import pytest

from conftest import ShellTest
from test_helper import KEY_ACTION, deleteIfExists


def set_up_search(test) -> None:
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "kuzu is cool" AS b;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "the shell is fun" AS c;\r')
    assert test.shell_process.expect_exact(["\u2502 the shell is fun \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.ENTER.value,
        "\n",
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_enter(test, key) -> None:
    set_up_search(test)
    test.send_statement("databases")
    test.send_finished_statement(key)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_R.value,
        KEY_ACTION.CTRL_N.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_search_next_once(test, key, history_path) -> None:
    # test search next once
    set_up_search(test)
    test.send_control_statement(key)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 the shell is fun \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_R.value,
        KEY_ACTION.CTRL_N.value,
    ],
)
@pytest.mark.parametrize('mode', [':multiline\n', ':singleline\n'])
def test_search_next_top(temp_db, mode, key, history_path) -> None:
    # test search next until top of history
    test = ShellTest().add_argument(temp_db).add_argument("-p").add_argument(history_path)
    test.start()
    test.send_finished_statement('RETURN "databases are really cool" AS a;\r')
    test.send_finished_statement(mode)
    set_up_search(test)
    for _ in range(6):
        test.send_control_statement(key)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases are really cool \u2502", pexpect.EOF]) == 0
    deleteIfExists(os.path.join(history_path, "history.txt"))


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_S.value,
        KEY_ACTION.CTRL_P.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_search_prev_once(test, key) -> None:
    # test search prev once
    set_up_search(test)
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_control_statement(key)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 the shell is fun \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_S.value,
        KEY_ACTION.CTRL_P.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_search_prev_bottom(test, key) -> None:
    # test search prev until bottom of history
    test.send_finished_statement('RETURN "kuzu is cool" AS b;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "the shell is fun" AS c;\r')
    assert test.shell_process.expect_exact(["\u2502 the shell is fun \u2502", pexpect.EOF]) == 0
    test.send_statement('RETURN "searching history" AS d;')
    # start search and move further into history
    for _ in range(3):
        test.send_control_statement(KEY_ACTION.CTRL_R.value)
    # move back to bottom of history
    for _ in range(3):
        test.send_control_statement(key)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 searching history \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_a(test) -> None:
    test.send_finished_statement('"databases rule" AS a;\r')
    assert (
        test.shell_process.expect_exact(
            [
                "Error: Parser exception: extraneous input '\"databases rule\"'",
                pexpect.EOF,
            ],
        )
        == 0
    )
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("databases")
    test.send_control_statement(KEY_ACTION.CTRL_A.value)
    test.send_statement("RETURN ")
    test.send_statement("\x1b[F") # move cursor to end
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_tab(test) -> None:
    # test tab becomes space when part of pasted input
    set_up_search(test)
    test.send_statement("databases\trule")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_tab_complete(test) -> None:
    # test tab complete search and go to end of line
    test.send_finished_statement('RETURN databases;\r')
    assert (
        test.shell_process.expect_exact(
            ['Error: Binder exception: Variable databases is not in scope.', pexpect.EOF],
        )
        == 0
    )
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("databases")
    test.send_statement(KEY_ACTION.TAB.value)
    test.send_statement(KEY_ACTION.BACKSPACE.value)  # remove semicolon
    test.send_statement("\x1b[D" * 9)
    test.send_statement('"\x1b[F')
    test.send_finished_statement(' rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_e(test) -> None:
    test.send_finished_statement('RETURN ;\r')
    assert (
        test.shell_process.expect_exact(
            ['Error: Parser exception: Invalid input <RETURN ;>: expected rule oC_RegularQuery (line: 1, offset: 7)',
             pexpect.EOF],
        )
        == 0
    )
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("return")
    test.send_control_statement(KEY_ACTION.CTRL_E.value)
    test.send_statement(KEY_ACTION.BACKSPACE.value)  # remove semicolon
    test.send_finished_statement('"databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_b(test) -> None:
    test.send_finished_statement('RETURN "databasesrule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databasesrule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("databasesr")
    test.send_control_statement(KEY_ACTION.CTRL_B.value)
    test.send_finished_statement(" \x1b[F\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_f(test) -> None:
    test.send_finished_statement('RETURN "databasesrule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databasesrule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("database")
    test.send_control_statement(KEY_ACTION.CTRL_F.value)
    test.send_finished_statement(" \x1b[F\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_t(test) -> None:
    test.send_finished_statement('RETURN "database srule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 database srule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("database ")
    test.send_control_statement(KEY_ACTION.CTRL_T.value)
    test.send_finished_statement("\x1b[F\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_u_cleared(test) -> None:
    # check if line was cleared
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("databases")
    test.send_control_statement(KEY_ACTION.CTRL_U.value)
    test.send_finished_statement('RETURN "kuzu is cool" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_u_position(test) -> None:
    # check if position in history was maintained
    set_up_search(test)
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("is cool")
    test.send_control_statement(KEY_ACTION.CTRL_U.value)
    test.send_finished_statement("\x1b[A\r")  # move up in history
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_k(test) -> None:
    test.send_finished_statement('RETURN "databases rule" AS aabc;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("As a")
    test.send_control_statement(KEY_ACTION.CTRL_K.value)
    test.send_finished_statement(";\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_d_eof(test) -> None:
    # ctrl d accepts search and acts as EOF when line is empty
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_control_statement(KEY_ACTION.CTRL_D.value)
    # search gets accepted
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0
    # acts as EOF
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 1


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_d_delete(test) -> None:
    # test accepts search and delete
    test.send_finished_statement('RETURN "databases/ rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases/ rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("databases")
    test.send_control_statement(KEY_ACTION.CTRL_D.value)
    test.send_finished_statement("\x1b[F\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_l_empty(test) -> None:
    # clear screen with empty line
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_control_statement(KEY_ACTION.CTRL_L.value)
    assert test.shell_process.expect_exact(["\x1b[H\x1b[2J", pexpect.EOF]) == 0
    test.send_statement("databases")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_l_non_empty(test) -> None:
    # clear screen with non empty line
    test.send_statement('RETURN "databases rule" AS a;')
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_control_statement(KEY_ACTION.CTRL_L.value)
    assert test.shell_process.expect_exact(["\x1b[H\x1b[2J", pexpect.EOF]) == 0
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_C.value,
        KEY_ACTION.CTRL_G.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_cancel_search_cleared(test, key) -> None:
    # check if line is cleared
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("databases")
    test.send_control_statement(key)
    test.send_finished_statement('RETURN "kuzu is cool" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_C.value,
        KEY_ACTION.CTRL_G.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_cancel_search_position(test, key) -> None:
    # check if position in history was maintained
    set_up_search(test)
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("is cool")
    test.send_control_statement(KEY_ACTION.CTRL_U.value)
    test.send_statement("\x1b[B")  # move down in history
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_H.value,
        KEY_ACTION.BACKSPACE.value,
        KEY_ACTION.CTRL_W.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_backspace(test, key) -> None:
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_R.value)
    test.send_statement("databasesabc")
    test.send_statement(key * 3)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0



import sys
# arg 1 is input Cypher.g4
inputCypher = open(sys.argv[1], 'r')
# arg 2 is input keywords.txt
inputKeywords = [x[:-1] for x in open(sys.argv[2], 'r')]
# arg 3 is output Cypher.g4
outputCypher = open(sys.argv[3], 'w')
# arg 4 is output keywords.h
outputKeywords = open(sys.argv[4], 'w')

print("Generating keywords for Cypher.g4")
antlr4String = ''.join(map(
    lambda x: x + ' : ' + ' '.join([f'( \'{i.upper()}\' | \'{i.lower()}\' )' if i != '_' else f'\'{i}\'' for i in x]) + ' ;\n\n',
    inputKeywords))

cypherPrefixString = '''
/*
* OpenCypher grammar at "https://s3.amazonaws.com/artifacts.opencypher.org/legacy/Cypher.g4"
*/
grammar Cypher;

// provide ad-hoc error messages for common syntax errors
@parser::declarations {
    virtual void notifyQueryNotConcludeWithReturn(antlr4::Token* startToken) {};
    virtual void notifyNodePatternWithoutParentheses(std::string nodeName, antlr4::Token* startToken) {};
    virtual void notifyInvalidNotEqualOperator(antlr4::Token* startToken) {};
    virtual void notifyEmptyToken(antlr4::Token* startToken) {};
    virtual void notifyReturnNotAtEnd(antlr4::Token* startToken) {};
    virtual void notifyNonBinaryComparison(antlr4::Token* startToken) {};
}

'''

outputCypher.write(cypherPrefixString + antlr4String + '\n' + ''.join(inputCypher))
print("keywords for Cypher.g4 generated")

print("Generating keywordList for keywords.h")
asCList = '{"' + '", "'.join(inputKeywords) + '"}'

outputKeywords.write(f'''#ifndef _keywordList
// clang-format off
#define _keywordList {asCList}
#define _keywordListLength {len(inputKeywords)}
// clang-format on
#endif
''')
print("keywordList generated")



from pathlib import Path

import kuzu
import pytest
from test_helper import KUZU_ROOT


def get_used_page_ranges(conn, table, column=None):
    used_pages = []
    if column is None:
        storage_info = conn.execute(f'call storage_info("{table}") return start_page_idx, num_pages')
    else:
        storage_info = conn.execute(
            f'call storage_info("{table}") where prefix(column_name, "{column}") return start_page_idx, num_pages'
        )
    while storage_info.has_next():
        cur_tuple = storage_info.get_next()
        # don't add empty (e.g. constant-compressed) pages
        if cur_tuple[1] > 0:
            used_pages.append(cur_tuple)
    return used_pages


def get_free_page_ranges(conn):
    free_pages = []
    result = conn.execute("call fsm_info() return *")
    while result.has_next():
        cur_tuple = result.get_next()
        assert cur_tuple[1] > 0
        free_pages.append(cur_tuple)
    return free_pages


def combine_adjacent_page_ranges(page_ranges):
    new_page_ranges = []
    last_range = None
    for page_range in page_ranges:
        if last_range is not None and last_range[0] + last_range[1] == page_range[0]:
            new_range = [last_range[0], last_range[1] + page_range[1]]
            last_range = new_range
        else:
            if last_range is not None:
                new_page_ranges.append(last_range)
            last_range = page_range
    if last_range:
        new_page_ranges.append(last_range)
    return new_page_ranges


def compare_page_range_lists(used_list, free_list):
    used_list.sort()
    free_list.sort()
    assert combine_adjacent_page_ranges(used_list) == free_list


@pytest.fixture
def fsm_node_table_setup(tmp_path: Path):
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    conn.execute("call threads=1")
    conn.execute("call auto_checkpoint=false")
    conn.execute(
        "create node table person (ID INt64, fName StRING, gender INT64, isStudent BoOLEAN, isWorker BOOLEAN, age INT64, eyeSight DOUBLE, birthdate DATE, registerTime TIMESTAMP, lastJobDuration interval, workedHours INT64[], usedNames STRING[], courseScoresPerTerm INT64[][], grades INT64[4], height float, u UUID, PRIMARY KEY (ID));"
    )
    conn.execute(
        f"COPY person FROM ['{KUZU_ROOT}/dataset/tinysnb/vPerson.csv', '{KUZU_ROOT}/dataset/tinysnb/vPerson2.csv'](ignore_errors=true, header=false)"
    )
    return db, conn


@pytest.fixture
def fsm_rel_table_setup(fsm_node_table_setup):
    _, conn = fsm_node_table_setup
    conn.execute(
        "create rel table knows (FROM person TO person, date DATE, meetTime TIMESTAMP, validInterval INTERVAL, comments STRING[], summary STRUCT(locations STRING[], transfer STRUCT(day DATE, amount INT64[])), notes UNION(firstmet DATE, type INT16, comment STRING), someMap MAP(STRING, STRING), MANY_MAnY);"
    )
    conn.execute(
        f"COPY knows FROM ['{KUZU_ROOT}/dataset/tinysnb/eKnows.csv', '{KUZU_ROOT}/dataset/tinysnb/eKnows_2.csv']"
    )
    return fsm_node_table_setup


@pytest.fixture
def fsm_rel_group_setup(tmp_path: Path):
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    conn.execute("call threads=1")
    conn.execute("call auto_checkpoint=false")
    conn.execute("create node table personA (ID INt64, fName StRING, PRIMARY KEY (ID));")
    conn.execute("create node table personB (ID INt64, fName StRING, PRIMARY KEY (ID));")
    conn.execute("create rel table likes (FROM personA TO personB, FROM personB To personA, date DATE);")
    conn.execute(f'COPY personA FROM "{KUZU_ROOT}/dataset/rel-group/node.csv";')
    conn.execute(f'COPY personB FROM "{KUZU_ROOT}/dataset/rel-group/node.csv";')
    conn.execute(f'COPY likes_personA_personB FROM "{KUZU_ROOT}/dataset/rel-group/edge.csv";')
    conn.execute(f'COPY likes_personB_personA FROM "{KUZU_ROOT}/dataset/rel-group/edge.csv";')
    return db, conn


def test_fsm_reclaim_list_column(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person", "workedHours")
    conn.execute("alter table person drop workedHours")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_string_column(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person", "fName")
    conn.execute("alter table person drop fName")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_list_of_lists(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person", "courseScoresPerTerm")
    conn.execute("alter table person drop courseScoresPerTerm")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_node_table(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person")
    conn.execute("drop table person")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)

def test_fsm_reclaim_node_table_delete(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person")
    conn.execute("match (p:person) delete p")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)

def test_fsm_reclaim_rel_table(fsm_rel_table_setup) -> None:
    _, conn = fsm_rel_table_setup
    used_pages = get_used_page_ranges(conn, "knows")
    conn.execute("drop table knows")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)

def test_fsm_reclaim_rel_table_delete(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person")
    conn.execute("match (p:person) delete p")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)

def test_fsm_reclaim_struct(fsm_rel_table_setup) -> None:
    _, conn = fsm_rel_table_setup
    used_pages = get_used_page_ranges(conn, "knows", "fwd_summary") + get_used_page_ranges(
        conn, "knows", "bwd_summary"
    )
    conn.execute("alter table knows drop summary")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_rel_group(fsm_rel_group_setup) -> None:
    _, conn = fsm_rel_group_setup
    used_pages = get_used_page_ranges(conn, "likes_personA_personB") + get_used_page_ranges(
        conn, "likes_personB_personA"
    )
    conn.execute("drop table likes")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_rel_group_column(fsm_rel_group_setup) -> None:
    _, conn = fsm_rel_group_setup
    used_pages = (
        get_used_page_ranges(conn, "likes_personA_personB", "fwd_date")
        + get_used_page_ranges(conn, "likes_personA_personB", "bwd_date")
        + get_used_page_ranges(conn, "likes_personB_personA", "fwd_date")
        + get_used_page_ranges(conn, "likes_personB_personA", "bwd_date")
    )
    conn.execute("alter table likes drop date")
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)



import os

import pexpect
import pytest
from conftest import ShellTest
from test_helper import KEY_ACTION, deleteIfExists


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_C.value,
        KEY_ACTION.CTRL_G.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_sigint_two_consecutive(key, test) -> None:
    # test two consecutive signit required to exit shell
    test.send_control_statement(key)
    test.send_control_statement(key)
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 1


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_C.value,
        KEY_ACTION.CTRL_G.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_sigint_line_interupts(key, test) -> None:
    # test sending line interupts the two signits to quit
    test.send_control_statement(key)
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(key)
    test.send_control_statement(key)
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 1


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_C.value,
        KEY_ACTION.CTRL_G.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_sigint_cancels_query(key, test) -> None:
    # test sigint cancels the query
    test.send_statement("CREATE NODE TABLE t0(i STRING, PRIMARY KEY(i));")
    test.send_control_statement(key)
    test.send_finished_statement("MATCH (a:t0) RETURN *;\r")
    assert (
        test.shell_process.expect_exact(
            ["Error: Binder exception: Table t0 does not exist.", pexpect.EOF],
        )
        == 0
    )


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.ENTER.value,
        "\n",
    ],
)
@pytest.mark.parametrize('mode', [':multiline\n', ':singleline\n'])
def test_enter(temp_db, key, history_path, mode) -> None:
    test = ShellTest().add_argument(temp_db).add_argument("-p").add_argument(history_path)
    test.start()
    test.send_finished_statement(mode)
    test.send_statement('RETURN "databases rule" AS a;')
    test.send_finished_statement(key)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0

    with open(os.path.join(history_path, "history.txt")) as f:
        assert f.readline() == mode
        assert f.readline() == 'RETURN "databases rule" AS a;\n'
    deleteIfExists(os.path.join(history_path, "history.txt"))

    # enter in multiline when not at the end of the line should not execute the query
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument("-p")
        .add_argument(history_path)
    )
    test.start()
    test.send_finished_statement(mode)
    test.send_statement('RETURN "databases rule" AS a;\x1b[D\x1b[D\x1b[D')
    test.send_finished_statement(key)
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0

    with open(os.path.join(history_path, "history.txt")) as f:
        assert f.readline() == mode
        assert f.readline() != 'RETURN "databases rule" AS a;\n'
    deleteIfExists(os.path.join(history_path, "history.txt"))


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.BACKSPACE.value,
        KEY_ACTION.CTRL_H.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_backspace(key, test) -> None:
    test.send_statement('RETURN "databases rule" AS a;abc')
    test.send_statement(key * 3)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_d_eof(test) -> None:
    # ctrl d acts as EOF when line is empty
    test.send_control_statement(KEY_ACTION.CTRL_D.value)
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 1


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_d_delete(test) -> None:
    # ctrl d deletes characters to the right
    test.send_statement('RETURN "databases rule" AS a;abc')
    test.send_statement("\x1b[D" * 3)  # move cursor to the left
    for _ in range(4):
        test.send_control_statement(KEY_ACTION.CTRL_D.value)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_t(test) -> None:
    test.send_statement('RETURN "databases rule" AS ;a')
    test.send_statement("\x1b[D")  # move cursor to the left
    test.send_control_statement(KEY_ACTION.CTRL_T.value)
    test.shell_process.sendcontrol(KEY_ACTION.CTRL_T.value)  # line does not refresh
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_b(test) -> None:
    test.send_statement(";")
    test.send_control_statement(KEY_ACTION.CTRL_B.value)
    test.shell_process.sendcontrol(KEY_ACTION.CTRL_B.value)  # line does not refresh
    test.send_finished_statement('RETURN "databases rule" AS a\x1b[C\r') # move cursor to the right one
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_f(test) -> None:
    test.send_statement('RETURN "databases rule" AS a')
    test.send_statement("\x1b[D")  # move cursor to the left
    test.send_control_statement(KEY_ACTION.CTRL_F.value)
    test.shell_process.sendcontrol(KEY_ACTION.CTRL_F.value)  # line does not refresh
    test.send_finished_statement(";\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_p(test) -> None:
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_P.value)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_n(test) -> None:
    test.send_finished_statement('RETURN "kuzu is cool" AS b;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_statement("\x1b[A" * 2)  # move up in history
    test.send_control_statement(KEY_ACTION.CTRL_N.value)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_R.value,
        KEY_ACTION.CTRL_S.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_search_opens(test, key, history_path) -> None:
    # test search opens
    test.send_control_statement(key)
    assert test.shell_process.expect_exact(["kuzu", pexpect.EOF]) == 0
    assert test.shell_process.expect_exact(["bck-i-search: _", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_R.value,
        KEY_ACTION.CTRL_S.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_search_good(test, key) -> None:
    # test successful search
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "kuzu is cool" AS b;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_control_statement(key)
    test.send_statement("databases")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_R.value,
        KEY_ACTION.CTRL_S.value,
    ],
)
@pytest.mark.parametrize('mode', [':multiline\n', ':singleline\n'])
def test_search_bad(mode, temp_db, key, history_path) -> None:
    # test failing search
    test = ShellTest().add_argument(temp_db).add_argument("-p").add_argument(history_path)
    test.start()
    test.send_finished_statement(mode)
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "kuzu is cool" AS b;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_control_statement(key)
    test.send_statement("databases are cool")
    assert test.shell_process.expect_exact(["failing-bck-i-search:", pexpect.EOF]) == 0
    # enter should process last match
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    deleteIfExists(os.path.join(history_path, "history.txt"))


@pytest.mark.parametrize(
    "key",
    [
        KEY_ACTION.CTRL_R.value,
        KEY_ACTION.CTRL_S.value,
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_search(test, key) -> None:
    # test starting search with text inputted already
    test.send_statement('RETURN "databases rule" AS a;')
    test.send_control_statement(key)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_u(test) -> None:
    test.send_statement("CREATE NODE TABLE t0(i STRING, PRIMARY KEY(i));")
    test.send_control_statement(KEY_ACTION.CTRL_U.value)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    test.send_finished_statement("MATCH (a:t0) RETURN *;\r")
    assert (
        test.shell_process.expect_exact(
            ["Error: Binder exception: Table t0 does not exist.", pexpect.EOF],
        )
        == 0
    )


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_k(test) -> None:
    test.send_statement('RETURN "databases rule" AS a;abc')
    test.send_statement("\x1b[D" * 3)  # move cursor to the left
    test.send_control_statement(KEY_ACTION.CTRL_K.value)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_a(test) -> None:
    test.send_statement('ETURN "databases rule" AS a;')
    test.send_control_statement(KEY_ACTION.CTRL_A.value)
    test.send_finished_statement("R\x1b[F\r") # move cursor to end of line
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_e(test) -> None:
    test.send_statement('RETURN "databases rule" AS a')
    test.send_statement("\x1b[H")  # move cursor to the front
    test.send_control_statement(KEY_ACTION.CTRL_E.value)
    test.send_finished_statement(";\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_l_empty(test) -> None:
    # clear screen with empty line
    test.send_control_statement(KEY_ACTION.CTRL_L.value)
    assert test.shell_process.expect_exact(["\x1b[H\x1b[2J", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_l_non_empty(test) -> None:
    # clear screen with non empty line
    test.send_statement('RETURN "databases rule" AS a')
    test.send_control_statement(KEY_ACTION.CTRL_L.value)
    assert test.shell_process.expect_exact(["\x1b[H\x1b[2J", pexpect.EOF]) == 0
    test.send_finished_statement(";\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_ctrl_w(test) -> None:
    test.send_statement('RETURN "databases rule" AS a; abc def ghi ')
    test.send_statement(KEY_ACTION.CTRL_W.value * 3)
    test.send_statement("\x1b[H")  # move cursor to the front
    test.send_statement(KEY_ACTION.CTRL_W.value)
    test.send_statement("\x1b[F")  # move cursor to the end
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_tab_spaces(test) -> None:
    # tab is interpreted as four spaces when pasting input
    test.send_finished_statement('RETURN "databases\trule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases    rule \u2502", pexpect.EOF]) == 0

@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_tab(test) -> None:
    # tab completion
    test.send_statement("CRE")
    test.send_statement(KEY_ACTION.TAB.value)
    test.send_statement(" N")
    test.send_statement(KEY_ACTION.TAB.value)
    test.send_statement(KEY_ACTION.TAB.value)
    test.send_statement(" TAB")
    test.send_statement(KEY_ACTION.TAB.value)
    test.send_statement(KEY_ACTION.TAB.value)
    test.send_statement("LE t0(i STRING, PRIMARY KEY(i));")
    test.send_statement(KEY_ACTION.TAB.value)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert (
        test.shell_process.expect_exact(
            ["\u2502 Table t0 has been created. \u2502", pexpect.EOF],
        )
        == 0
    )



from __future__ import annotations

import asyncio
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Any

from .connection import Connection
from .prepared_statement import PreparedStatement

if TYPE_CHECKING:
    import sys
    from types import TracebackType

    from .database import Database
    from .query_result import QueryResult

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


class AsyncConnection:
    """AsyncConnection enables asynchronous execution of queries with a pool of connections and threads."""

    def __init__(
        self,
        database: Database,
        max_concurrent_queries: int = 4,
        max_threads_per_query: int = 0,
    ) -> None:
        """
        Initialise the async connection.

        Parameters
        ----------
        database : Database
            Database to connect to.

        max_concurrent_queries : int
            Maximum number of concurrent queries to execute. This corresponds to the
            number of connections and thread pool size. Default is 4.

        max_threads_per_query : int
            Controls the maximum number of threads per connection that can be used
            to execute one query. Default is 0, which means no limit.
        """
        self.database = database
        self.connections = [Connection(database) for _ in range(max_concurrent_queries)]
        self.connections_counter = [0 for _ in range(max_concurrent_queries)]
        self.lock = threading.Lock()

        for conn in self.connections:
            conn.init_connection()
            conn.set_max_threads_for_exec(max_threads_per_query)

        self.executor = ThreadPoolExecutor(max_workers=max_concurrent_queries)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    def __get_connection_with_least_queries(self) -> tuple[Connection, int]:
        with self.lock:
            conn_index = self.connections_counter.index(min(self.connections_counter))
            self.connections_counter[conn_index] += 1
        return self.connections[conn_index], conn_index

    def __decrement_connection_counter(self, conn_index: int) -> None:
        """Decrement the query counter for a connection."""
        with self.lock:
            self.connections_counter[conn_index] -= 1
            if self.connections_counter[conn_index] < 0:
                self.connections_counter[conn_index] = 0

    def acquire_connection(self) -> Connection:
        """
        Acquire a connection from the connection pool for temporary synchronous
        calls. If the connection pool is oversubscribed, the method will return
        the connection with the least number of queued queries. It is required
        to release the connection by calling `release_connection` after the
        connection is no longer needed.

        Returns
        -------
        Connection
            A connection object.
        """
        conn, _ = self.__get_connection_with_least_queries()
        return conn

    def release_connection(self, conn: Connection) -> None:
        """
        Release a connection acquired by `acquire_connection` back to the
        connection pool. Calling this method is required when the connection is
        no longer needed.

        Parameters
        ----------
        conn : Connection
            Connection object to release.


        """
        for i, existing_conn in enumerate(self.connections):
            if existing_conn == conn:
                self.__decrement_connection_counter(i)
                break

    def set_query_timeout(self, timeout_in_ms: int) -> None:
        """
        Set the query timeout value in ms for executing queries.

        Parameters
        ----------
        timeout_in_ms : int
            query timeout value in ms for executing queries.

        """
        for conn in self.connections:
            conn.set_query_timeout(timeout_in_ms)

    async def execute(
        self, query: str | PreparedStatement, parameters: dict[str, Any] | None = None
    ) -> QueryResult | list[QueryResult]:
        """
        Execute a query asynchronously.

        Parameters
        ----------
        query : str | PreparedStatement
            A prepared statement or a query string.
            If a query string is given, a prepared statement will be created
            automatically.

        parameters : dict[str, Any]
            Parameters for the query.

        Returns
        -------
        QueryResult
            Query result.

        """
        loop = asyncio.get_running_loop()
        # If the query is a prepared statement, use the connection associated with it
        if isinstance(query, PreparedStatement):
            conn = query._connection
            for i, existing_conn in enumerate(self.connections):
                if existing_conn == conn:
                    conn_index = i
                    with self.lock:
                        self.connections_counter[conn_index] += 1
                    break
        else:
            conn, conn_index = self.__get_connection_with_least_queries()

        try:
            return await loop.run_in_executor(self.executor, conn.execute, query, parameters)
        except asyncio.CancelledError:
            conn.interrupt()
        finally:
            self.__decrement_connection_counter(conn_index)

    async def prepare(self, query: str) -> PreparedStatement:
        """
        Create a prepared statement for a query asynchronously.

        Parameters
        ----------
        query : str
            Query to prepare.

        Returns
        -------
        PreparedStatement
            Prepared statement.

        """
        loop = asyncio.get_running_loop()
        conn, conn_index = self.__get_connection_with_least_queries()

        try:
            preparedStatement = await loop.run_in_executor(self.executor, conn.prepare, query)
            return preparedStatement
        finally:
            self.__decrement_connection_counter(conn_index)

    def close(self) -> None:
        """
        Close all connections and shutdown the thread pool.

        Note: Call to this method is optional. The connections and thread pool
        will be closed automatically when the instance is garbage collected.
        """
        for conn in self.connections:
            conn.close()

        self.executor.shutdown(wait=True)



import asyncio
import time

import kuzu
import pytest


@pytest.mark.asyncio()
async def test_async_prepare_and_execute(async_connection_readonly):
    query = "MATCH (a:person) WHERE a.ID = $1 RETURN a.age;"
    prepared = await async_connection_readonly.prepare(query)
    result = await async_connection_readonly.execute(prepared, {"1": 0})
    assert result.has_next()
    assert result.get_next() == [35]
    assert not result.has_next()
    result.close()
    for i in async_connection_readonly.connections_counter:
        assert i == 0


@pytest.mark.asyncio()
async def test_async_prepare_and_execute_concurrent(async_connection_readonly):
    num_queries = 100
    query = "RETURN $1;"
    prepared_statements = await asyncio.gather(*[async_connection_readonly.prepare(query) for _ in range(num_queries)])
    results = await asyncio.gather(*[
        async_connection_readonly.execute(prepared, {"1": i}) for i, prepared in enumerate(prepared_statements)
    ])
    for i, result in enumerate(results):
        assert result.has_next()
        assert result.get_next() == [i]
        assert not result.has_next()
        result.close()
    for i in async_connection_readonly.connections_counter:
        assert i == 0


@pytest.mark.asyncio()
async def test_async_query(async_connection_readonly):
    query = "MATCH (a:person) WHERE a.ID = 0 RETURN a.age;"
    result = await async_connection_readonly.execute(query)
    assert result.has_next()
    assert result.get_next() == [35]
    assert not result.has_next()
    result.close()

    query = "MATCH (a:person) WHERE a.ID = 2 RETURN a.age;"
    result = await async_connection_readonly.execute(query)
    assert result.has_next()
    assert result.get_next() == [30]
    assert not result.has_next()
    result.close()
    for i in async_connection_readonly.connections_counter:
        assert i == 0


@pytest.mark.asyncio()
async def test_async_query_concurrent(async_connection_readonly):
    num_queries = 100
    queries = [f"RETURN {i};" for i in range(num_queries)]
    results = await asyncio.gather(*[async_connection_readonly.execute(query) for query in queries])
    for i, result in enumerate(results):
        assert result.has_next()
        assert result.get_next() == [i]
        assert not result.has_next()
        result.close()
    for i in async_connection_readonly.connections_counter:
        assert i == 0


@pytest.mark.asyncio()
async def test_async_query_multiple_results(async_connection_readonly):
    query = "MATCH (a:person) WHERE a.ID = 0 RETURN a.age; MATCH (a:person) WHERE a.ID = 2 RETURN a.age;"
    results = await async_connection_readonly.execute(query)
    assert len(results) == 2
    result = results[0]
    assert result.has_next()
    assert result.get_next() == [35]
    assert not result.has_next()

    result = results[1]
    assert result.has_next()
    assert result.get_next() == [30]
    assert not result.has_next()
    for result in results:
        result.close()
    for i in async_connection_readonly.connections_counter:
        assert i == 0


@pytest.mark.asyncio()
async def test_async_connection_create_and_close():
    db = kuzu.Database(":memory:", buffer_pool_size=2**28)
    async_connection = kuzu.AsyncConnection(db)
    for _ in range(10):
        res = await async_connection.execute("RETURN 1;")
        assert res.has_next()
        assert res.get_next() == [1]
        assert not res.has_next()
        res.close()
    num_queries = 100
    queries = [f"RETURN {i};" for i in range(num_queries)]
    results = await asyncio.gather(*[async_connection.execute(query) for query in queries])
    for i, result in enumerate(results):
        assert result.has_next()
        assert result.get_next() == [i]
        assert not result.has_next()
        result.close()
    async_connection.close()
    db.close()


def test_acquire_connection(async_connection_readonly):
    conn = async_connection_readonly.acquire_connection()
    assert conn is not None
    assert async_connection_readonly.connections_counter[0] == 1
    result = conn.execute("RETURN 1;")
    assert result.has_next()
    assert result.get_next() == [1]
    assert not result.has_next()
    result.close()
    async_connection_readonly.release_connection(conn)
    for i in async_connection_readonly.connections_counter:
        assert i == 0


@pytest.mark.asyncio()
async def test_async_connection_interrupt(async_connection_readonly) -> None:
    query = "UNWIND RANGE(1,1000000) AS x UNWIND RANGE(1, 1000000) AS y RETURN COUNT(x + y);"
    async_connection_readonly.set_query_timeout(100 * 1000)
    prepared = await async_connection_readonly.prepare(query)
    task = asyncio.create_task(async_connection_readonly.execute(prepared))
    time.sleep(5)
    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task
